'''
    author: rkopeinig
    script: Basic Statistics and Trend Analysis
    description: Bunch of basic statistics and trend analysis algorithms.
                 See Proba-V Time-Series Analysis Toolbox Poster.
    version: 0.1
'''

import numpy as np
import pandas as pd
import statsmodels.api as sm
from scipy.stats import norm
from scipy import interpolate


def mann_kendall(series, alpha=0.05):
    """ Kendall's tau coefficient to measure the ordinal association between two measured quantities

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param alpha: Significance level
        :type alpha: integer or float

        :return param trend: Returns if trend is increasing, decreasing or there is no trend
        :rtype: string

        :return param h: Returns if trend is present or absence
        :rtype: boolean

        :return param p: P value of the significance test
        :rtype: float

        :return param z: Normalized test statistics
        :rtype: float

    """
    n = len(series)
    s = 0
    for k in range(n - 1):
        for j in range(k + 1, n):
            s += np.sign(series[j] - series[k])

    unique_x = np.unique(series)
    g = len(unique_x)

    if n == g:  # no tie
        var_s = (n * (n - 1) * (2 * n + 5)) / 18
    else:  # some ties in data
        tp = np.zeros(unique_x.shape)
        for i in range(len(unique_x)):
            tp[i] = sum(unique_x[i] == series)
        var_s = (n * (n - 1) * (2 * n + 5) + np.sum(tp * (tp - 1) * (2 * tp + 5))) / 18

    if s > 0:
        z = (s - 1) / np.sqrt(var_s)
    elif s == 0:
        z = 0
    elif s < 0:
        z = (s + 1) / np.sqrt(var_s)

    p = 2 * (1 - norm.cdf(abs(z)))  # two tail test
    h = abs(z) > norm.ppf(1 - alpha / 2)

    if (z < 0) and h:
        trend = 'decreasing'
    elif (z > 0) and h:
        trend = 'increasing'
    else:
        trend = 'no trend'
    return trend, h, p, z


def temporal_smoothing_pandas(series, method='cubic'):
    """ Temporal Smoothing using Cubic and Spline

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param method: Smoothing Method either cubic or spline. Cubic is set as default
        :type method: string

        :return: Pandas Series of smoothed Time-Series using cubic and spline interpolation methods.
        :rtype: pandas.core.series.Series

    """
    if method=='cubic':
        return series.interpolate(method='cubic')
    if method=='spline':
        return series.interpolate(method='spline', order=3, s=0)

def temporal_smoothing_scipy(series):
    """ Temporal Smoothing using Spline

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :return: Pandas Series of smoothed Time-Series using Scipy's spline interpolation methods.
        :rtype: pandas.core.series.Series

    """
    x = range(1, len(series) + 1)
    tck = interpolate.splrep(x, series.values)
    return pd.Series(interpolate.splev(x, tck),index=series.index)


def seasonal_decomposition(series, freq=30, model='additive'):
    """ Seasonal Decompositon

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param freq: Frequency of the series. Must be used if x is not a pandas object.
        :type freq: integer

        :param model: Type of seasonal component. Abbreviations are accepted.
        :type model: string

        :return param df: DataFrame of seasonal decomposition results including seasonal, trend, and resid attributes.
        :rtype param df: pandas.DataFrame

    """
    sd = sm.tsa.seasonal_decompose(series, model=model, freq=freq)
    return pd.DataFrame({'observed': sd.observed.values, 'resid': sd.resid.values,
                       'seasonal': sd.seasonal.values, 'trend': sd.trend.values},
                      index=sd.observed.index)


def de_trender(series, order=1):
    """ Perform a de-trending on a Pandas Series.

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param order: Specifies the polynomial order of the trend, zero is constant, one is linear trend, two is quadratic trend
        :type order: integer

        :return: Pandas Series of input data and de-trended data.
        :rtype: pandas.core.series.Series

    """
    detrend = sm.tsa.tsatools.detrend(series.values, order=order)
    return pd.Series(detrend, index=series.index)


def peak_signals(series):
    """ Find the peaks in Pandas Series.

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :return: Pandas Series of found peaks (Boolean).
        :rtype: pandas.core.series.Series

    """
    i=np.array(series.values)
    signal = (i>np.roll(i,1)) & (i>np.roll(i,-1))
    return pd.Series(signal, index=series.index)

def descriptive_statistics(df):
    """ Descriptive Statistic

        :param df: Two-dimensional size-mutable ndarray with axis labels including Time-Series.
        :type df: pandas.core.series.Series

        :return param df: DataFrame of descriptive statistics about given Pandas DataFrame.
        :rtype param df: pandas.DataFrame

    """
    grouped_data = df.groupby(['data'])
    return grouped_data['data'].describe().unstack()


def pivot_table(df, columns='year', index='month'):
    """ Pivot Table

        :param df: Two-dimensional size-mutable ndarray with axis labels including Time-Series.
        :type df: pandas.core.series.Series

        :return: DataFrame Pivot Table with yearly, monthly or daily attributes.
        :rtype: pandas.DataFrame.pivot

    """
    df['year'] = df.index.year
    df['month'] = df.index.month
    df['day'] = df.index.day
    return pd.pivot_table(df, values='data', columns=columns, index=index)